from functools import cached_property
from typing import List
from typing import Optional
from typing import Union

import requests
from pydantic import BaseModel
from pydantic import BaseSettings

from anaconda_cloud_auth import __version__ as version

OIDC_REQUEST_HEADERS = {"User-Agent": f"anaconda-cloud-auth/{version}"}


class APIConfig(BaseSettings):
    class Config:
        env_prefix = "ANACONDA_CLOUD_API_"
        env_file = ".env"
        keep_untouched = (
            cached_property,  # this allows pydantic models to have a cached_property
        )

    domain: str = "anaconda.cloud"
    key: Optional[str] = None

    @cached_property
    def aau_token(self) -> Union[str, None]:
        # The token is cached in anaconda_anon_usage, so we can also cache here
        try:
            from anaconda_anon_usage.tokens import token_string
        except ImportError:
            return None

        try:
            return token_string()
        except Exception:
            # We don't want this to block user login in any case,
            # so let any Exceptions pass silently.
            return None


class AuthConfig(BaseSettings):
    class Config:
        env_prefix = "ANACONDA_CLOUD_AUTH_"
        env_file = ".env"

    domain: str = "id.anaconda.cloud"
    client_id: str = "b4ad7f1d-c784-46b5-a9fe-106e50441f5a"
    redirect_uri: str = "http://127.0.0.1:8000/auth/oidc"
    openid_config_path: str = ".well-known/openid-configuration"

    @property
    def well_known_url(self: "AuthConfig") -> str:
        """The URL from which to load the OpenID configuration."""
        return f"https://{self.domain}/{self.openid_config_path}"

    @property
    def oidc(self) -> "OpenIDConfiguration":
        """The OIDC configuration, cached as a regular instance attribute."""
        res = requests.get(self.well_known_url, headers=OIDC_REQUEST_HEADERS)
        res.raise_for_status()
        oidc_config = OpenIDConfiguration(**res.json())
        return self.__dict__.setdefault("_oidc", oidc_config)


class OpenIDConfiguration(BaseModel):
    authorization_endpoint: str
    token_endpoint: str
    jwks_uri: str

    id_token_signing_alg_values_supported: List[str] = []



import time
from multiprocessing.pool import ThreadPool
from threading import Thread
from typing import Iterator

import pytest
import requests

from anaconda_cloud_auth.exceptions import AuthenticationError
from anaconda_cloud_auth.handlers import AuthCodeRedirectServer
from anaconda_cloud_auth.handlers import capture_auth_code
from anaconda_cloud_auth.handlers import shutdown_all_servers


@pytest.fixture()
def server() -> Iterator[AuthCodeRedirectServer]:
    """Start up the web server responsible for capturing the auth code in a background thread."""
    oidc_path = "/auth/oidc"
    host_name = "localhost"
    server_port = 8080

    server = AuthCodeRedirectServer(oidc_path, (host_name, server_port))

    def _f() -> None:
        with server:
            server.handle_request()

    t = Thread(target=_f, daemon=True)
    t.start()
    yield server


def test_server_response_success(server: AuthCodeRedirectServer) -> None:
    """The server captures the query parameters and then redirects to the success page."""
    # Make the request and ensure the code is captured by the server
    response = requests.get(
        "http://localhost:8080/auth/oidc?code=something&state=some-state"
    )
    assert server.result is not None
    assert server.result.auth_code == "something"
    assert server.result.state == "some-state"

    assert response.status_code == 200
    assert response.url == "https://anaconda.cloud/local-login-success"


@pytest.mark.parametrize(
    "query_params",
    [
        pytest.param("state=some-state", id="missing-code"),
        pytest.param("code=something", id="missing-state"),
    ],
)
def test_server_response_error(
    server: AuthCodeRedirectServer, query_params: str
) -> None:
    """We redirect to the error page if we forget the code or state parameters."""
    response = requests.get(
        f"http://localhost:8080/auth/oidc?state=some-state?{query_params}"
    )
    assert response.status_code == 200
    assert response.url == "https://anaconda.cloud/local-login-error"


def test_server_response_not_found(server: AuthCodeRedirectServer) -> None:
    """Return a 404 if the path is not the OIDC path."""
    response = requests.get(
        "http://localhost:8080/auth/oidc2?code=some-code&state=some-state"
    )
    assert response.status_code == 404


def test_shutdown_server_before_completing_authentication() -> None:
    """If e.g. Navigator user starts the auth flow, but tries to close Navigator before finishing, the
    shutdown_all_servers() function can be used to prevent deadlock."""
    with ThreadPool(processes=1) as pool:
        # Start a server in a background thread to capture the auth code
        async_result = pool.apply_async(
            capture_auth_code, ("http://localhost:8000", "random-state")
        )

        # We need to sleep a bit before we can shut down the server
        time.sleep(0.1)

        # Shut it down from the main thread
        shutdown_all_servers()

        # We will have an authentication error because we never captured the code
        with pytest.raises(AuthenticationError):
            async_result.get()

    # If we get this far, we have not hit a deadlock
    assert True



import os
from collections import defaultdict
from pathlib import Path
from typing import Any
from typing import Union

import pytest
from _pytest.monkeypatch import MonkeyPatch
from dotenv import load_dotenv
from keyring.backend import KeyringBackend
from pytest_mock import MockerFixture

from anaconda_cloud_auth import config
from anaconda_cloud_auth import login
from anaconda_cloud_auth.client import BaseClient
from anaconda_cloud_auth.console import console
from anaconda_cloud_auth.token import TokenInfo

load_dotenv()


class MockedKeyring(KeyringBackend):
    """A high-priority in-memory keyring backend for testing"""

    priority = 10000.0  # type: ignore
    _data: dict = defaultdict(dict)

    def __init__(self) -> None:
        super().__init__()

    def set_password(self, service: str, username: str, password: str) -> None:
        self._data[service][username] = password

    def get_password(self, service: str, username: str) -> Union[str, None]:
        password = self._data.get(service, {}).get(username, None)
        return password

    def delete_password(self, service: str, username: str) -> None:
        _ = self._data.get(service, {}).pop(username)


@pytest.fixture(autouse=True)
def clear_mocked_keyring() -> None:
    MockedKeyring._data = defaultdict(dict)


@pytest.fixture(autouse=True)
def set_keyring_name(mocker: MockerFixture) -> None:
    mocker.patch("anaconda_cloud_auth.token.KEYRING_NAME", "Anaconda Cloud Test")


@pytest.fixture
def outdated_token_info() -> TokenInfo:
    # This is an old token from the dev system that will always be out-of-date
    api_key = (
        "eyJhbGciOiJSUzI1NiIsImtpZCI6ImRlZmF1bHQiLCJ0eXAiOiJKV1QifQ"
        ".eyJleHAiOjE2ODkwODg3ODYsInN1YiI6ImQwNGMzNTZiLWFmZDItNGIzZ"
        "S04MGYyLTQwMzExM2UwOTc0YiJ9.tTi_gttpQWhiTy_Uh0bDohN34mqd_6"
        "AHvyXf8_R5PFxjI-z9Ei0S3XCm9siP0RfyJx2j08SRs3FwXpkT8b8jP__C"
        "h-Y4K-zXYksZnTGcQ77YhKQCoKpGSpGlE4yD6gRXRRDT7bHs4H7gf4e6iD"
        "1Vdcq0yx5-5h-CbBgSwS9LSpJ_HDZBUy-xbRrw0aD36aQ5qs6huswgCOQa"
        "9YrYfsrSbZW8uY48LAt5Y69t8x1twNBI5_Cumx-JEZuDLQxq7HQp7wKldE"
        "tbycV5uemKjyR1Qeuva2zCKYB3FEXdTEiWHhTzhSQ-3-xjUrIZvpfGJd3G"
        "CzXlkUhpeDoj2KbSN-Lq0Q"
    )
    return TokenInfo(api_key=api_key, domain="mocked-domain")


@pytest.fixture()
def tmp_cwd(monkeypatch: MonkeyPatch, tmp_path: Path) -> Path:
    """Create & return a temporary directory after setting current working directory to it."""
    monkeypatch.chdir(tmp_path)
    return tmp_path


@pytest.fixture(scope="session")
def is_not_none() -> Any:
    """
    An object that can be used to test whether another is None.

    This is particularly useful when testing contents of collections, e.g.:

    ```python
    def test_data(data, is_not_none):
        assert data == {"some_key": is_not_none, "some_other_key": 5}
    ```

    """

    class _NotNone:
        def __eq__(self, other: Any) -> bool:
            return other is not None

    return _NotNone()


@pytest.fixture
def disable_dot_env(monkeypatch: MonkeyPatch) -> None:
    from anaconda_cloud_auth.config import APIConfig
    from anaconda_cloud_auth.config import AuthConfig

    monkeypatch.setattr(APIConfig.Config, "env_file", "")
    monkeypatch.setattr(AuthConfig.Config, "env_file", "")


@pytest.fixture()
def integration_test_client(monkeypatch: MonkeyPatch) -> BaseClient:
    """Provides a request client configured to talk to the automation environment.

    We first load credentials from environment variables. We then use a special
    user credential to generate an API key via basic login. Finally, we patch
    the configuration and clients such that CloudFlare headers are included and
    the domains match the automation environment.

    """
    if (email := os.getenv("TEST_AUTOMATION_USER_EMAIL")) is None:
        raise ValueError(
            "TEST_AUTOMATION_USER_EMAIL must be specified as an environment variable or in `.env`"
        )
    if (password := os.getenv("TEST_AUTOMATION_USER_PASSWORD")) is None:
        raise ValueError(
            "TEST_AUTOMATION_USER_PASSWORD must be specified as an environment variable or in `.env`"
        )

    monkeypatch.setenv(
        "ANACONDA_CLOUD_AUTH_DOMAIN", "nucleus-automation.anacondaconnect.com/api/iam"
    )
    monkeypatch.setenv(
        "ANACONDA_CLOUD_API_DOMAIN", "nucleus-automation.anacondaconnect.com"
    )
    monkeypatch.setenv(
        "ANACONDA_CLOUD_AUTH_CLIENT_ID", "e0648d7e-72c1-4159-b7e8-5d020ac134c2"
    )

    # Here, we store Cloudflare headers to allow use of Warp from GitHub Actions runners.
    # These are stored as secrets in Vault.
    if client_id := os.getenv("CF_CLIENT_ID"):
        monkeypatch.setitem(
            config.OIDC_REQUEST_HEADERS, "CF-Access-Client-Id", client_id
        )
    if client_secret := os.getenv("CF_CLIENT_SECRET"):
        monkeypatch.setitem(
            config.OIDC_REQUEST_HEADERS, "CF-Access-Client-Secret", client_secret
        )

    def mock_input(msg: str, **kwargs: Any) -> str:
        """Mock the input function to mimic user entry."""
        assert email is not None
        assert password is not None
        if msg == "Please enter your email: ":
            return email
        elif msg == "Please enter your password: ":
            return password
        else:
            raise ValueError(f"Unknown input statement: {msg}")

    monkeypatch.setattr(console, "input", mock_input)

    login(basic=True)

    client = BaseClient()
    if client_id:
        client.headers["CF-Access-Client-Id"] = client_id
    if client_secret:
        client.headers["CF-Access-Client-Secret"] = client_secret

    return client


def pytest_addoption(parser):  # type: ignore
    """Defines custom CLI options."""
    parser.addoption(
        "--integration",
        action="store_true",
        dest="integration",
        default=False,
        help="enable integration tests",
    )


def pytest_collection_modifyitems(config, items):  # type: ignore
    """Auto-mark each test in the integration directory, and enable them based on CLI flag."""
    integration_test_root_dir = Path(__file__).parent / "integration"
    run_integration_tests = config.getoption("--integration")
    for item in items:
        # Here, we add a marker to any test in the "tests/integration" directory
        if integration_test_root_dir in Path(item.fspath).parents:
            item.add_marker(pytest.mark.integration)

        # Add a skip marker if the CLI option is not used. We use an additional marker so that we can
        # independently select integrations with `pytest -m integration` and enable them with `--integration`.
        if "integration" in item.keywords and not run_integration_tests:
            item.add_marker(pytest.mark.skip(reason="need --integration to run"))


@pytest.fixture
def with_aau_token(mocker: MockerFixture) -> None:
    mocker.patch("anaconda_cloud_auth.config.APIConfig.aau_token", "anon-token")


@pytest.fixture
def without_aau_token(mocker: MockerFixture) -> None:
    mocker.patch("anaconda_cloud_auth.config.APIConfig.aau_token", None)



from pytest_mock import MockerFixture

from anaconda_cloud_auth.config import AuthConfig


def test_legacy() -> None:
    config = AuthConfig(domain="anaconda.cloud/api/iam")
    assert config.oidc.authorization_endpoint == "https://anaconda.cloud/authorize"
    assert config.oidc.token_endpoint == "https://anaconda.cloud/api/iam/token"


def test_well_known_headers(mocker: MockerFixture) -> None:
    import requests

    spy = mocker.spy(requests, "get")

    config = AuthConfig()
    assert config.oidc
    spy.assert_called_once()
    assert (
        spy.call_args.kwargs.get("headers", {})
        .get("User-Agent")
        .startswith("anaconda-cloud-auth")
    )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

from anaconda_cloud_auth import __version__
from anaconda_cloud_auth import login
from anaconda_cloud_auth.actions import _get_api_key
from anaconda_cloud_auth.client import BaseClient
from anaconda_cloud_auth.config import AuthConfig
from anaconda_cloud_auth.token import TokenInfo

if TYPE_CHECKING:
    from _pytest.monkeypatch import MonkeyPatch


@pytest.fixture(autouse=True)
def set_dev_env_vars(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv(
        "ANACONDA_CLOUD_API_DOMAIN", "nucleus-latest.anacondaconnect.com"
    )
    monkeypatch.setenv("ANACONDA_CLOUD_AUTH_DOMAIN", "dev.id.anaconda.cloud")
    monkeypatch.setenv(
        "ANACONDA_CLOUD_AUTH_CLIENT_ID", "83d245e3-6312-4f44-9298-1f5b32a13769"
    )


@pytest.mark.integration
@pytest.mark.usefixtures("integration_test_client")
def test_login_to_token_info(is_not_none: Any) -> None:
    auth_config = AuthConfig()
    keyring_token = TokenInfo.load(auth_config.domain)

    assert keyring_token == {
        "domain": auth_config.domain,
        "username": None,
        "api_key": is_not_none,
    }


@pytest.mark.integration
def test_get_auth_info(integration_test_client: BaseClient, is_not_none: Any) -> None:
    response = integration_test_client.get("/api/account")
    assert response.status_code == 200
    assert response.json() == {
        "user": is_not_none,
        "profile": is_not_none,
        "subscriptions": is_not_none,
    }


@pytest.fixture
def mocked_do_login(mocker: MockerFixture) -> MagicMock:
    def _mocked_login(auth_config: AuthConfig, basic: bool) -> None:
        TokenInfo(domain=auth_config.domain, api_key="from-login").save()

    mocker.patch("anaconda_cloud_auth.actions._do_login", _mocked_login)
    from anaconda_cloud_auth import actions

    login_spy = mocker.spy(actions, "_do_login")
    return login_spy


def test_login_no_existing_token(mocked_do_login: MagicMock) -> None:
    auth_config = AuthConfig()
    login(auth_config=auth_config)

    assert TokenInfo.load(auth_config.domain).api_key == "from-login"
    mocked_do_login.assert_called_once()


def test_login_has_valid_token(
    mocked_do_login: MagicMock, mocker: MockerFixture
) -> None:
    auth_config = AuthConfig()

    mocker.patch("anaconda_cloud_auth.token.TokenInfo.expired", False)
    TokenInfo(domain=auth_config.domain, api_key="pre-existing").save()

    login(auth_config=auth_config)
    mocked_do_login.assert_not_called()

    assert TokenInfo.load(auth_config.domain).api_key == "pre-existing"


def test_force_login_with_valid_token(
    mocked_do_login: MagicMock, mocker: MockerFixture
) -> None:
    auth_config = AuthConfig()

    mocker.patch("anaconda_cloud_auth.token.TokenInfo.expired", False)
    TokenInfo(domain=auth_config.domain, api_key="pre-existing").save()

    login(auth_config=auth_config, force=True)
    mocked_do_login.assert_called_once()

    assert TokenInfo.load(auth_config.domain).api_key == "from-login"


def test_login_has_expired_token(
    mocked_do_login: MagicMock, mocker: MockerFixture
) -> None:
    auth_config = AuthConfig()

    mocker.patch("anaconda_cloud_auth.token.TokenInfo.expired", True)
    TokenInfo(domain=auth_config.domain, api_key="pre-existing-expired").save()

    login(auth_config=auth_config)
    mocked_do_login.assert_called_once()

    assert TokenInfo.load(auth_config.domain).api_key == "from-login"


class MockResponse:
    def __init__(self, json_data: dict[str, Any], status_code: int):
        self.json_data = json_data
        self.status_code = status_code

    def json(self) -> dict[str, Any]:
        return self.json_data


class MockedRequest:
    def __init__(self) -> None:
        self.called_with_args: tuple[Any] = ()  # type: ignore
        self.called_with_kwargs: dict[str, Any] = {}

    def __call__(self, *args: Any, **kwargs: Any) -> MockResponse:
        self.called_with_args = args  # type: ignore
        self.called_with_kwargs = kwargs
        return MockResponse(status_code=201, json_data={"api_key": "some-jwt"})


@pytest.fixture()
def mocked_request(mocker: MockerFixture) -> MockedRequest:
    """A mocked post request returning an API key."""

    # This could be generalized further, but it may not be worth the effort
    # For now, this mimics a fixed POST request with fixed mocked return data

    mocked_request = MockedRequest()
    mocker.patch("requests.post", mocked_request)
    return mocked_request


@pytest.mark.usefixtures("without_aau_token")
def test_get_api_key(mocked_request: MockedRequest) -> None:
    """When we get an API key, we assign appropriate generic scopes and tags."""

    key = _get_api_key("some_access_token")
    assert key == "some-jwt"

    headers = mocked_request.called_with_kwargs["headers"]
    assert headers["Authorization"].startswith("Bearer")
    assert "X-AAU-CLIENT" not in headers

    data = mocked_request.called_with_kwargs["json"]
    assert data == {
        "scopes": ["cloud:read", "cloud:write"],
        "tags": [f"anaconda-cloud-auth/v{__version__}"],
    }


@pytest.mark.usefixtures("with_aau_token")
def test_get_api_key_with_aau_token(mocked_request: MockedRequest) -> None:
    """When we get an API key, we assign appropriate generic scopes and tags."""

    key = _get_api_key("some_access_token")
    assert key == "some-jwt"

    headers = mocked_request.called_with_kwargs["headers"]
    assert headers["Authorization"].startswith("Bearer")
    assert headers["X-AAU-CLIENT"] == "anon-token"



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from pytest_mock import MockerFixture

from anaconda_cloud_auth.client import BaseClient
from anaconda_cloud_auth.client import client_factory
from anaconda_cloud_auth.exceptions import LoginRequiredError
from anaconda_cloud_auth.token import TokenInfo

if TYPE_CHECKING:
    from _pytest.monkeypatch import MonkeyPatch


def test_login_required_error() -> None:
    client = BaseClient()
    with pytest.raises(LoginRequiredError):
        _ = client.get("/api/account")


def test_user_agent_client_factory() -> None:
    client = client_factory("my-app/version")
    response = client.get("/api/catalogs/examples")
    assert response.request.headers.get("User-Agent") == "my-app/version"


def test_anonymous_endpoint(monkeypatch: MonkeyPatch, disable_dot_env: None) -> None:
    _ = disable_dot_env
    monkeypatch.setenv("ANACONDA_CLOUD_API_DOMAIN", "anaconda.cloud")
    monkeypatch.setenv("ANACONDA_CLOUD_AUTH_DOMAIN", "dummy")
    monkeypatch.delenv("ANACONDA_CLOUD_API_KEY", raising=False)

    client = BaseClient()
    response = client.get("/api/catalogs/examples")
    assert "Authorization" not in response.request.headers.keys()
    assert response.status_code == 200


def test_token_included(
    mocker: MockerFixture,
    monkeypatch: MonkeyPatch,
    outdated_token_info: TokenInfo,
    disable_dot_env: None,
) -> None:
    _ = disable_dot_env
    monkeypatch.setenv("ANACONDA_CLOUD_AUTH_DOMAIN", "mocked-domain")
    mocker.patch("anaconda_cloud_auth.token.TokenInfo.expired", False)
    monkeypatch.delenv("ANACONDA_CLOUD_API_KEY", raising=False)

    outdated_token_info.save()

    client = BaseClient()
    response = client.get("/api/catalogs/examples")
    assert (
        response.request.headers.get("Authorization")
        == f"Bearer {outdated_token_info.api_key}"
    )


def test_api_key_env_variable_over_keyring(
    outdated_token_info: TokenInfo, monkeypatch: MonkeyPatch
) -> None:
    outdated_token_info.save()
    monkeypatch.setenv("ANACONDA_CLOUD_API_KEY", "set-in-env")

    client = BaseClient()
    assert client.config.key == "set-in-env"

    response = client.get("/api/catalogs/examples")
    assert response.request.headers.get("Authorization") == "Bearer set-in-env"


def test_api_key_init_arg_over_variable(
    outdated_token_info: TokenInfo, monkeypatch: MonkeyPatch
) -> None:
    outdated_token_info.save()
    monkeypatch.setenv("ANACONDA_CLOUD_API_KEY", "set-in-env")

    client = BaseClient(api_key="set-in-init")
    assert client.config.key == "set-in-init"

    response = client.get("/api/catalogs/examples")
    assert response.request.headers.get("Authorization") == "Bearer set-in-init"



import platform
from pathlib import Path

import pytest
from keyring.errors import PasswordDeleteError
from pytest_mock import MockerFixture

from anaconda_cloud_auth.actions import logout
from anaconda_cloud_auth.config import AuthConfig
from anaconda_cloud_auth.token import TokenExpiredError
from anaconda_cloud_auth.token import TokenInfo
from anaconda_cloud_auth.token import TokenNotFoundError


def test_expired_token_error(outdated_token_info: TokenInfo) -> None:
    with pytest.raises(TokenExpiredError):
        _ = outdated_token_info.get_access_token()


def test_token_not_found() -> None:
    auth_config = AuthConfig()

    with pytest.raises(TokenNotFoundError):
        _ = TokenInfo.load(auth_config.domain)

    with pytest.raises(TokenNotFoundError):
        _ = TokenInfo(domain=auth_config.domain).get_access_token()


def test_logout_multiple_okay(mocker: MockerFixture) -> None:
    """We can logout multiple times and no exception is raised."""
    import keyring

    delete_spy = mocker.spy(keyring, "delete_password")

    auth_config = AuthConfig(domain="test")
    token_info = TokenInfo(api_key="key", domain=auth_config.domain)
    token_info.save()

    for _ in range(2):
        logout(auth_config)

    delete_spy.assert_called_once()


def test_anaconda_keyring_priority() -> None:
    import keyring.backend

    backends = {k.name: k for k in keyring.backend.get_all_keyring()}

    assert "token AnacondaKeyring" in backends
    assert backends["token AnacondaKeyring"].priority == 9
    assert (
        backends["token AnacondaKeyring"].priority
        < backends["chainer ChainerBackend"].priority
    )


def test_anaconda_keyring_save_delete(tmp_path: Path) -> None:
    from anaconda_cloud_auth.token import AnacondaKeyring

    fn = tmp_path / "keyring"
    AnacondaKeyring.keyring_path = fn
    assert AnacondaKeyring.viable

    anaconda_keyring = AnacondaKeyring()

    assert anaconda_keyring.get_password("s", "u") is None

    with pytest.raises(PasswordDeleteError):
        anaconda_keyring.delete_password("s", "u")

    anaconda_keyring.set_password("s", "u", "p")
    assert anaconda_keyring.keyring_path.exists()
    assert anaconda_keyring.keyring_path.read_text() == '{"s": {"u": "p"}}'

    anaconda_keyring.set_password("s", "u2", "p")
    assert anaconda_keyring.keyring_path.read_text() == '{"s": {"u": "p", "u2": "p"}}'

    assert anaconda_keyring.viable

    assert anaconda_keyring.get_password("s", "u") == "p"
    assert anaconda_keyring.get_password("s", "u3") is None

    anaconda_keyring.delete_password("s", "u")
    assert anaconda_keyring.keyring_path.read_text() == '{"s": {"u2": "p"}}'

    anaconda_keyring.set_password("s2", "u", "p")
    assert (
        anaconda_keyring.keyring_path.read_text()
        == '{"s": {"u2": "p"}, "s2": {"u": "p"}}'
    )

    anaconda_keyring.delete_password("s", "u2")
    assert anaconda_keyring.keyring_path.read_text() == '{"s": {}, "s2": {"u": "p"}}'

    with pytest.raises(PasswordDeleteError):
        anaconda_keyring.delete_password("s", "u2")

    assert anaconda_keyring.get_password("s3", "u4") is None


def test_anaconda_keyring_empty(tmp_path: Path) -> None:
    fn = tmp_path / "keyring"
    fn.touch()
    assert fn.exists()

    from anaconda_cloud_auth.token import AnacondaKeyring

    AnacondaKeyring.keyring_path = fn

    anaconda_keyring = AnacondaKeyring()
    assert anaconda_keyring.get_password("s", "u") is None

    with pytest.raises(PasswordDeleteError):
        anaconda_keyring.delete_password("s", "u")


@pytest.mark.skipif(
    platform.system() == "Windows", reason="This has been hard to test in CI/CD"
)
def test_anaconda_keyring_not_writable() -> None:
    from anaconda_cloud_auth.token import AnacondaKeyring

    if platform.system() == "Windows":
        root = Path("c:\\windows")
    else:
        root = Path("/root")
    AnacondaKeyring.keyring_path = root / "keyring"

    assert not AnacondaKeyring.viable
